/*
Copyright © 2018 Roesner P/L

Dual com port serial controller.

baud rates:	bredal=9600 
			rds=4800 
			amatron=19200/57600 
			teejet = 9600 
			ZURF = 9600

*/
//___________________________________________________________________________________

var SerialPort = require('serialport');
var fs = require('fs');
var Gpio = require('pigpio').Gpio;

//log to console if it exists, else set console.log to nop
if (console){
	console.log('Console log enabled');
}
else {
	console.log = function(){};
}

//aux led is dual colour, turn it off
var ledRed = new Gpio(2, {mode: Gpio.OUTPUT});
var ledGreen = new Gpio(3, {mode: Gpio.OUTPUT});
ledRed.digitalWrite(0);
ledGreen.digitalWrite(0);

var hardwareVer = '10G.025';
var softwareVer = '10.04.07';
var saveConfigTimer;

var appRateCommand = 0;
var port0TxCounts = 0;
var port0RxCounts = 0;
var port1TxCounts = 0;
var port1RxCounts = 0;

var port0Mode;
var port0Protocol;
var port0Baud;
var port0DataBits;
var port0Parity;
var port0StopBits;
var port0FlowCtrl;
var port0DispConfig;
var port1Mode;
var port1Protocol;
var port1Baud;
var port1DataBits;
var port1Parity;
var port1StopBits;
var port1FlowCtrl;
var port1DispConfig;

//start by reading the config file
readConfigFile();


function readConfigFile(){
	try {
		//read and parse the config file
		var rawData = fs.readFileSync('/home/pi/servermain/config.json');
		var config = JSON.parse(rawData);
		
		port0Mode = config.port0Mode;
		port0Protocol = config.port0Protocol;
		port0Baud = config.port0Baud;
		port0DataBits = config.port0DataBits;
		port0Parity = config.port0Parity;
		port0StopBits = config.port0StopBits;
		port0FlowCtrl = config.port0FlowCtrl;
		port0DispConfig = config.port0DispConfig;
		port1Mode = config.port1Mode;
		port1Protocol = config.port1Protocol;
		port1Baud = config.port1Baud;
		port1DataBits = config.port1DataBits;
		port1Parity = config.port1Parity;
		port1StopBits = config.port1StopBits;
		port1FlowCtrl = config.port1FlowCtrl;
		port1DispConfig = config.port1DispConfig;

		console.log(config);
	}
	catch (e){
		//config file is blank or faulty, so load the factory data
		console.log('Catch error readConfigFile');
		loadFactoryConfig();
	}
}

function loadFactoryConfig(){
	try {
		//populate factory config data
		port0Mode = 'Rate Out';
		port0Protocol = 'LH5000';
		port0Baud = 9600;
		port0DataBits = 8;
		port0Parity = 'none';
		port0StopBits = 1;
		port0FlowCtrl = false;
		port0DispConfig = '8N1';
		port1Mode = 'Off';
		port1Protocol = 'LH5000';
		port1Baud = 9600;
		port1DataBits = 8;
		port1Parity = 'none';
		port1StopBits = 1;
		port1FlowCtrl = false;
		port1DispConfig = '8N1';
		
		console.log('Loading factory config');
	}
	catch (e){
		console.log('Catch error loadFactoryConfig');
	}
}

function saveConfigFile(){
	try {
		//load variables
		var configData = {
			port0Mode: port0Mode,
			port0Protocol: port0Protocol,
			port0Baud: port0Baud,
			port0DataBits: port0DataBits,
			port0Parity: port0Parity,
			port0StopBits: port0StopBits,
			port0FlowCtrl: port0FlowCtrl,
			port0DispConfig: port0DispConfig,
			port1Mode: port1Mode,
			port1Protocol: port1Protocol,
			port1Baud: port1Baud,
			port1DataBits: port1DataBits,
			port1Parity: port1Parity,
			port1StopBits: port1StopBits,
			port1FlowCtrl: port1FlowCtrl,
			port1DispConfig: port1DispConfig
		};
		
		//convert to json strings and write to file
		var data = JSON.stringify(configData, null, 2);
		fs.writeFileSync('/home/pi/servermain/config.json', data);
		
		console.log('Saving config file');
	}
	catch (e){
		console.log('Catch error saveConfigFile');
	}
}


//set the port options
var port0 = new SerialPort('/dev/ttyACM0', {
	baudRate: port0Baud,
	dataBits: port0DataBits,
	parity: port0Parity,
	stopBits: port0StopBits,
	rtscts: port0FlowCtrl
});

//open errors will be emitted as an error event
port0.on('error', function(err) {
	console.log('Error: ', err.message);
})

//receive data
port0.on('data', function (data) {
	port0RxCounts ++;
	console.log('Port0 received: ' + data);
});



var port1 = new SerialPort('/dev/ttyACM1', {
	baudRate: port1Baud,
	dataBits: port1DataBits,
	parity: port1Parity,
	stopBits: port1StopBits,
	rtscts: port1FlowCtrl
});

//open errors will be emitted as an error event
port1.on('error', function(err) {
	console.log('Error: ', err.message);
})

//receive data
port1.on('data', function (data) {
	port1RxCounts ++;
	console.log('Port1 received: ' + data);
});




function port0SendRate(newRate){
	try {
		port0TxCounts ++;
		
		if (port0Protocol == 'LH5000'){
			
			//example packet for a command of 125kg/ha
			//port0.write([0xD2,0x00,0x04,0x48,0xE8,0x01,0x00,0x07]);
			//first 2 bytes 0xD200 is the command for set rate
			//next byte 0x04 is the number of data bytes
			//the 4 data bytes 0x48E80100 is 125000 parsed as lsb first
			var rawRate = newRate * 1000;
			var packet = new Buffer(8, 0);
			
			//command
			packet[0] = 0xD2;
			packet[1] = 0x00;
			
			//data length
			packet[2] = 0x04;
			
			//data value lsb first
			packet[3] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[4] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[5] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[6] = rawRate & 0xFF;
			
			//checksum calc, sum all bytes
			var chkSumPointer = 0;
			var chkSumValue = 0;
			
			while (chkSumPointer < 7)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			
			//load checksum value, ignore 8bit overrun
			packet[7] = chkSumValue & 0xFF;
			
			//send
			port0.write(packet);
			console.log('Sending port0 LH5000 Rate: ' + newRate.toString(10));
		}
		else if (port0Protocol == 'LH5000x10'){
			//Same as LH5000 except rate multiplied by 10

			//example packet for a command of 125kg/ha
			//port0.write([0xD2,0x00,0x04,0x48,0xE8,0x01,0x00,0x07]);
			//first 2 bytes 0xD200 is the command for set rate
			//next byte 0x04 is the number of data bytes
			//the 4 data bytes 0x48E80100 is 125000 parsed as lsb first
			var rawRate = newRate * 1000 * 10;
			var packet = new Buffer(8, 0);
			
			//command
			packet[0] = 0xD2;
			packet[1] = 0x00;
			
			//data length
			packet[2] = 0x04;
			
			//data value lsb first
			packet[3] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[4] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[5] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[6] = rawRate & 0xFF;
			
			//checksum calc, sum all bytes
			var chkSumPointer = 0;
			var chkSumValue = 0;
			
			while (chkSumPointer < 7)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			
			//load checksum value, ignore 8bit overrun
			packet[7] = chkSumValue & 0xFF;
			
			//send
			port0.write(packet);
			console.log('Sending port0 LH5000 Rate: ' + newRate.toString(10));
		}
		else if (port0Protocol == 'LH5000/10'){
			//Same as LH5000 except rate divided by 10

			//example packet for a command of 125kg/ha
			//port0.write([0xD2,0x00,0x04,0x48,0xE8,0x01,0x00,0x07]);
			//first 2 bytes 0xD200 is the command for set rate
			//next byte 0x04 is the number of data bytes
			//the 4 data bytes 0x48E80100 is 125000 parsed as lsb first
			var rawRate = newRate * 1000 / 10;
			var packet = new Buffer(8, 0);
			
			//command
			packet[0] = 0xD2;
			packet[1] = 0x00;
			
			//data length
			packet[2] = 0x04;
			
			//data value lsb first
			packet[3] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[4] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[5] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[6] = rawRate & 0xFF;
			
			//checksum calc, sum all bytes
			var chkSumPointer = 0;
			var chkSumValue = 0;
			
			while (chkSumPointer < 7)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			
			//load checksum value, ignore 8bit overrun
			packet[7] = chkSumValue & 0xFF;
			
			//send
			port0.write(packet);
			console.log('Sending port0 LH5000 Rate: ' + newRate.toString(10));
		}
		else if (port0Protocol == 'ZURF'){
			//first convert the rate value to a string
			//early zurf can only do up to 999, so cap it to this
			var rateString;
			
			if (newRate > 999)
			{
				rateString = '999';
			}
			else
			{
				rateString = newRate.toString(10);
			}
			
			//if less than 3 characters, pad with zeros
			while (rateString.length < 3)
			{
				rateString = '0' + rateString;
			}
			
			var packet = new Buffer(8, 0);
			
			//zurf controller reads ascii characters, to make checksum calcs
			//easier we send the hex representation of the ascii characters
			// First byte is "{" or hex 7B
			// Final byte is "}" or hex 7D
			packet[0] = 0x7B;
			packet[7] = 0x7D;

			//ascii 2 character command for rate is SD
			packet[1] = 0x53;
			packet[2] = 0x44;
			
			
			//ascii rate value as hex
			packet[3] = Number(rateString.charCodeAt(0)).toString(10) & 0xFF;
			packet[4] = Number(rateString.charCodeAt(1)).toString(10) & 0xFF;
			packet[5] = Number(rateString.charCodeAt(2)).toString(10) & 0xFF;
			
			//checksum calc, xor of all bytes
			var chkSumValue = 0;
			
			chkSumValue = packet[1] ^ packet[2];
			chkSumValue = chkSumValue ^ packet[3];
			chkSumValue = chkSumValue ^ packet[4];
			chkSumValue = chkSumValue ^ packet[5];
			
			//reserved checksum values are forced to 0x55, see zurf docs
			if (chkSumValue == 0x00 || chkSumValue == 0x7B || chkSumValue == 0x7D)
			{
				chkSumValue = 0x55;
			}
			
			//load checksum value
			packet[6] = chkSumValue & 0xFF;
			
			//send
			port0.write(packet);
			console.log('Sending port0 ZURF Rate: ' + rateString);
		}
		else if (port0Protocol == 'ASCP'){
			//amatron ASCP needs the rate value in IEEE754 32bit floating point format
			//below converts to binary, splits into: sign + exponent + mantissa
			//then dumps big endian hex bytes into a 4 byte buffer

			var floatVar = newRate;
			var floatBuf = Buffer.alloc(4);
			floatBuf.writeFloatBE(floatVar, 0);

			console.log('buffer: ' + floatBuf[3].toString(16) + floatBuf[2].toString(16) + floatBuf[1].toString(16) + floatBuf[0].toString(16));


			var readVar = floatBuf.readFloatBE(0);
			console.log('readVar: ' + readVar);

			//this protocol also uses special control characters
			//stx = 0x02 start of text, identifies the start of the packet
			//eot = 0x04 end of transmission, identifies the end of the packet
			//dle = 0x10 data link escape, identifies the next byte as data
			//every special character within the packet requires a dle before it

			//populate the message, exlude all special control characters for now
			//multi byte values are sent least significant byte first
			var rawPacket = Buffer.alloc(11);

			rawPacket[0] = 0x00;
			rawPacket[1] = 0x01;
			rawPacket[2] = 0x07;
			rawPacket[3] = 0x05;
			rawPacket[4] = 0x00;
			rawPacket[5] = 0x00;
			rawPacket[6] = floatBuf[3];
			rawPacket[7] = floatBuf[2];
			rawPacket[8] = floatBuf[1];
			rawPacket[9] = floatBuf[0];

			//calc checksum and insert before the end of message byte
			var chkSumPointer = 0;
			var chkSumValue = 0;

			//sum all bytes, excl start of message, excl end of message
			while (chkSumPointer < 10){
				chkSumValue += rawPacket[chkSumPointer];
				chkSumPointer ++;
			}

			//ignore overrun of the sum, take two's complement
			rawPacket[10] = 0x100 - (chkSumValue & 0xFF);

			//special control values
			var stx = 0x02;
			var eot = 0x04;
			var dle = 0x10;

			//build a temp buffer incl byte stuffing, count the total bytes
			//if every byte excl start and end was a special character, we'd have 24bytes
			var stuffedPacket = Buffer.alloc(24);
			var rawPointer = 0;
			var byteVal = 0;
			var totalBytes = 0;

			//load start of message byte, does not require dle stuffing
			stuffedPacket[0] = stx;
			totalBytes ++;

			//load each raw packet byte
			while (rawPointer < 11){
				byteVal = rawPacket[rawPointer];
				
				//if it's a special character, insert a dle byte
				if (byteVal == stx || byteVal == eot || byteVal == dle){
					stuffedPacket[totalBytes] = dle;
					totalBytes ++;
				}
				
				//now insert the raw packet byte
				stuffedPacket[totalBytes] = rawPacket[rawPointer];
				totalBytes ++;
				
				rawPointer ++;
			}

			//load end of message byte, does not require dle stuffing
			stuffedPacket[totalBytes] = eot;
			totalBytes ++;

			//slice the temp buffer into the final packet to send
			var packet = stuffedPacket.slice(0, totalBytes);
			
			//send
			port0.write(packet);
			console.log('Sending port0 ASCP Rate: ' + newRate.toString(10));
			
			//console.log('total bytes: ' + totalBytes);
			//console.log('packet length: ' + packet.length);
			//console.log('packet bytes: ' + packet.toString('hex'));
		}
		else if (port0Protocol == 'VDSTAD'){
			//first convert the rate value to a string
			var rateString;
			
			//vaderstad can only do up to 999.0, so cap it to this
			if (newRate > 999)
			{
				rateString = '999.0';
			}
			//else force to one decimal place and convert to string
			else
			{
				newRate = newRate.toFixed(1);
				rateString = newRate.toString(10);
			}
			
			//console.log('rate string: ' + rateString);
			
			//15 is max possible chars required
			var packet = new Buffer(15, 0);
			
			//vaderstad controller reads ascii characters
			//start of frame is ascii $
			packet[0] = 0x24;
			//set rate command is ascii SET followed by comma
			packet[1] = 0x53;
			packet[2] = 0x45;
			packet[3] = 0x54;
			packet[4] = 0x2C;
			
			//the rate value will have a min of 3 bytes, example 1.0
			packet[5] = Number(rateString.charCodeAt(0)).toString(10) & 0xFF;
			packet[6] = Number(rateString.charCodeAt(1)).toString(10) & 0xFF;
			packet[7] = Number(rateString.charCodeAt(2)).toString(10) & 0xFF;
			
			//but possibly 5 bytes long, example 100.0
			var pointer = 7;
			
			if (rateString.length > 3){
				pointer ++;
				packet[pointer] = Number(rateString.charCodeAt(3)).toString(10) & 0xFF;
			}
			
			if (rateString.length > 4){
				pointer ++;
				packet[pointer] = Number(rateString.charCodeAt(4)).toString(10) & 0xFF;
			}
			
			//end of frame byte is ascii *
			pointer ++;
			packet[pointer] = 0x2A;
			
			//checksum calc, xor of all bytes
			//exclude start of frame byte, exclude end of frame byte
			var chkSumValue = 0;
			
			chkSumValue = packet[1] ^ packet[2];
			chkSumValue = chkSumValue ^ packet[3];
			chkSumValue = chkSumValue ^ packet[4];
			chkSumValue = chkSumValue ^ packet[5];
			chkSumValue = chkSumValue ^ packet[6];
			chkSumValue = chkSumValue ^ packet[7];
			
			if (rateString.length > 3){
				chkSumValue = chkSumValue ^ packet[8];
			}
			if (rateString.length > 4){
				chkSumValue = chkSumValue ^ packet[9];
			}
			
			//mask to one byte
			chkSumValue = chkSumValue & 0xFF;
			
			//now convert the hex byte to a string, and set letters to upper case
			chkSumValue = chkSumValue.toString(16);
			chkSumValue = chkSumValue.toUpperCase();
			//console.log('checksum hex string: ' + chkSumValue);
			
			//then send the hex character code for the first character as one byte
			pointer ++;
			packet[pointer] = Number(chkSumValue.charCodeAt(0));
			//console.log('ascii code for first checksum character: ' + packet[pointer].toString(16));
			
			//and the same for the second character
			pointer ++;
			packet[pointer] = Number(chkSumValue.charCodeAt(1));
			//console.log('ascii code for second checksum character: ' + packet[pointer].toString(16));
			
			
			//end of frame is carriage return then line feed
			pointer ++;
			packet[pointer] = 0x0D;
			pointer ++;
			packet[pointer] = 0x0A;
			
			//slice the packet, up to and including the pointer
			//this drops any unused bytes at the end of the buffer
			pointer ++;
			packet = packet.slice(0, pointer);
			
			//console.log(packet);
			//console.log('packet bytes: ' + packet.toString('hex'));
			
			//send
			port0.write(packet);
			console.log('Sending port0 VDSTAD Rate: ' + rateString);
		} 
		else if (port0Protocol == 'TEEJET844'){
			// Only difference between Teejet protocols is that the carrier application rate is limited to 999 for the 844 controller and to 9999 for the 855 controller.
			var rawRate;
                
			// Teejet 844 can only do up to 999, so cap it to this
			if (newRate > 999)
			{
				rawRate = 999;
			}
			// Only take integers, so remove any decimal places
			else
			{
				rawRate = newRate.toFixed(0);
			}
			var logRate = rawRate;
		
			// Packet is 15 bytes long
			// Bytes from position 0 to 11 are unused. Bytes in position 12 and 13 are rate. Byte in position 14 is checksum.
			// All are in Hexadecimal format.
			// When 2 bytes are used, the first received is the most significant byte.
		
			var packet = Buffer.alloc(15);
		
			packet[13] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[12] = rawRate & 0xFF;
		
		
			// Checksum is calculated by adding all bytes (without carry) and then substracting the result from 55 hexadecimal.
			var chkSumPointer = 0;
			var chkSumValue = 0;
		
			while (chkSumPointer < 14)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			chkSumValue = 0x55 - chkSumValue;
			packet[14] = chkSumValue & 0xFF;		
			

			//send
			port0.write(packet);
			console.log('Sending port0 TEEJET844 Rate: ' + logRate);
		}
		else if (port0Protocol == 'TEEJET855'){
			// Only difference between Teejet protocols is that the carrier application rate is limited to 999 for the 844 controller and to 9999 for the 855 controller.
			var rawRate;
                
			// Teejet 855 can only do up to 9999, so cap it to this
			if (newRate > 9999)
			{
				rawRate = 9999;
			}
			// Only take integers, so remove any decimal places
			else
			{
				rawRate = newRate.toFixed(0);
			}
			var logRate = rawRate;

		
			// Packet is 15 bytes long
			// Bytes from position 0 to 11 are unused. Bytes in position 12 and 13 are rate. Byte in position 14 is checksum.
			// All are in Hexadecimal format.
			// When 2 bytes are used, the first received is the most significant byte.
		
			var packet = Buffer.alloc(15);
		
			packet[13] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[12] = rawRate & 0xFF;
		
		
			// Checksum is calculated by adding all bytes (without carry) and then substracting the result from 55 hexadecimal.
			var chkSumPointer = 0;
			var chkSumValue = 0;
		
			while (chkSumPointer < 14)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			chkSumValue = 0x55 - chkSumValue;
			packet[14] = chkSumValue & 0xFF;		
			

			//send
			port0.write(packet);
			console.log('Sending port0 TEEJET855 Rate: ' + logRate);
		}
	}
	catch (e){
		console.log('Catch error port0SendRate');
	}
}


function port1SendRate(newRate){
	try {
		port1TxCounts ++;
		
		if (port1Protocol == 'LH5000'){
			//example packet for a command of 125kg/ha
			//port0.write([0xD2,0x00,0x04,0x48,0xE8,0x01,0x00,0x07]);
			//first 2 bytes 0xD200 is the command for set rate
			//next byte 0x04 is the number of data bytes
			//the 4 data bytes 0x48E80100 is 125000 parsed as lsb first
			var rawRate = newRate * 1000;
			var packet = new Buffer(8, 0);
			
			//command
			packet[0] = 0xD2;
			packet[1] = 0x00;
			
			//data length
			packet[2] = 0x04;
			
			//data value lsb first
			packet[3] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[4] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[5] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[6] = rawRate & 0xFF;
			
			//checksum calc, sum all bytes
			var chkSumPointer = 0;
			var chkSumValue = 0;
			
			while (chkSumPointer < 7)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			
			//load checksum value, ignore 8bit overrun
			packet[7] = chkSumValue & 0xFF;
			
			//send
			port1.write(packet);
			console.log('Sending port1 LH5000 Rate: ' + newRate.toString(10));
		}
		else if (port1Protocol == 'LH5000x10'){
			//Same as LH5000 except rate multiplied by 10

			//example packet for a command of 125kg/ha
			//port0.write([0xD2,0x00,0x04,0x48,0xE8,0x01,0x00,0x07]);
			//first 2 bytes 0xD200 is the command for set rate
			//next byte 0x04 is the number of data bytes
			//the 4 data bytes 0x48E80100 is 125000 parsed as lsb first
			var rawRate = newRate * 1000 * 10;
			var packet = new Buffer(8, 0);
			
			//command
			packet[0] = 0xD2;
			packet[1] = 0x00;
			
			//data length
			packet[2] = 0x04;
			
			//data value lsb first
			packet[3] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[4] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[5] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[6] = rawRate & 0xFF;
			
			//checksum calc, sum all bytes
			var chkSumPointer = 0;
			var chkSumValue = 0;
			
			while (chkSumPointer < 7)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			
			//load checksum value, ignore 8bit overrun
			packet[7] = chkSumValue & 0xFF;
			
			//send
			port1.write(packet);
			console.log('Sending port1 LH5000 Rate: ' + newRate.toString(10));
		}
		else if (port1Protocol == 'LH5000/10'){
			//Same as LH5000 except rate divided by 10

			//example packet for a command of 125kg/ha
			//port0.write([0xD2,0x00,0x04,0x48,0xE8,0x01,0x00,0x07]);
			//first 2 bytes 0xD200 is the command for set rate
			//next byte 0x04 is the number of data bytes
			//the 4 data bytes 0x48E80100 is 125000 parsed as lsb first
			var rawRate = newRate * 1000 / 10;
			var packet = new Buffer(8, 0);
			
			//command
			packet[0] = 0xD2;
			packet[1] = 0x00;
			
			//data length
			packet[2] = 0x04;
			
			//data value lsb first
			packet[3] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[4] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[5] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[6] = rawRate & 0xFF;
			
			//checksum calc, sum all bytes
			var chkSumPointer = 0;
			var chkSumValue = 0;
			
			while (chkSumPointer < 7)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			
			//load checksum value, ignore 8bit overrun
			packet[7] = chkSumValue & 0xFF;
			
			//send
			port1.write(packet);
			console.log('Sending port1 LH5000 Rate: ' + newRate.toString(10));
		}
		else if (port1Protocol == 'ZURF'){
			//first convert the rate value to a string
			//early zurf can only do up to 999, so cap it to this
			var rateString;
			
			if (newRate > 999)
			{
				rateString = '999';
			}
			else
			{
				rateString = newRate.toString(10);
			}
			
			//if less than 3 characters, pad with zeros
			while (rateString.length < 3)
			{
				rateString = '0' + rateString;
			}
			
			var packet = new Buffer(8, 0);
			
			//zurf controller reads ascii characters, to make checksum calcs
			//easier we send the hex representation of the ascii characters
			// First byte is "{" or hex 7B
			// Final byte is "}" or hex 7D
			packet[0] = 0x7B;
			packet[7] = 0x7D;

			//ascii 2 character command for rate is SD
			packet[1] = 0x53;
			packet[2] = 0x44;
			
			
			//ascii rate value as hex
			packet[3] = Number(rateString.charCodeAt(0)).toString(10) & 0xFF;
			packet[4] = Number(rateString.charCodeAt(1)).toString(10) & 0xFF;
			packet[5] = Number(rateString.charCodeAt(2)).toString(10) & 0xFF;
			
			//checksum calc, xor of all bytes
			var chkSumValue = 0;
			
			chkSumValue = packet[1] ^ packet[2];
			chkSumValue = chkSumValue ^ packet[3];
			chkSumValue = chkSumValue ^ packet[4];
			chkSumValue = chkSumValue ^ packet[5];
			
			//reserved checksum values are forced to 0x55, see zurf docs
			if (chkSumValue == 0x00 || chkSumValue == 0x7B || chkSumValue == 0x7D)
			{
				chkSumValue = 0x55;
			}
			
			//load checksum value
			packet[6] = chkSumValue & 0xFF;
			
			//send
			port1.write(packet);
			console.log('Sending port1 ZURF Rate: ' + rateString);
		}
		else if (port1Protocol == 'ASCP'){
			//amatron ASCP needs the rate value in IEEE754 32bit floating point format
			//below converts to binary, splits into: sign + exponent + mantissa
			//then dumps big endian hex bytes into a 4 byte buffer

			var floatVar = newRate;
			var floatBuf = Buffer.alloc(4);
			floatBuf.writeFloatBE(floatVar, 0);

			console.log('buffer: ' + floatBuf[3].toString(16) + floatBuf[2].toString(16) + floatBuf[1].toString(16) + floatBuf[0].toString(16));


			var readVar = floatBuf.readFloatBE(0);
			console.log('readVar: ' + readVar);

			//this protocol also uses special control characters
			//stx = 0x02 start of text, identifies the start of the packet
			//eot = 0x04 end of transmission, identifies the end of the packet
			//dle = 0x10 data link escape, identifies the next byte as data
			//every special character within the packet requires a dle before it

			//populate the message, exlude all special control characters for now
			//multi byte values are sent least significant byte first
			var rawPacket = Buffer.alloc(11);

			rawPacket[0] = 0x00;
			rawPacket[1] = 0x01;
			rawPacket[2] = 0x07;
			rawPacket[3] = 0x05;
			rawPacket[4] = 0x00;
			rawPacket[5] = 0x00;
			rawPacket[6] = floatBuf[3];
			rawPacket[7] = floatBuf[2];
			rawPacket[8] = floatBuf[1];
			rawPacket[9] = floatBuf[0];

			//calc checksum and insert before the end of message byte
			var chkSumPointer = 0;
			var chkSumValue = 0;

			//sum all bytes, excl start of message, excl end of message
			while (chkSumPointer < 10){
				chkSumValue += rawPacket[chkSumPointer];
				chkSumPointer ++;
			}

			//ignore overrun of the sum, take two's complement
			rawPacket[10] = 0x100 - (chkSumValue & 0xFF);

			//special control values
			var stx = 0x02;
			var eot = 0x04;
			var dle = 0x10;

			//build a temp buffer incl byte stuffing, count the total bytes
			//if every byte excl start and end was a special character, we'd have 24bytes
			var stuffedPacket = Buffer.alloc(24);
			var rawPointer = 0;
			var byteVal = 0;
			var totalBytes = 0;

			//load start of message byte, does not require dle stuffing
			stuffedPacket[0] = stx;
			totalBytes ++;

			//load each raw packet byte
			while (rawPointer < 11){
				byteVal = rawPacket[rawPointer];
				
				//if it's a special character, insert a dle byte
				if (byteVal == stx || byteVal == eot || byteVal == dle){
					stuffedPacket[totalBytes] = dle;
					totalBytes ++;
				}
				
				//now insert the raw packet byte
				stuffedPacket[totalBytes] = rawPacket[rawPointer];
				totalBytes ++;
				
				rawPointer ++;
			}

			//load end of message byte, does not require dle stuffing
			stuffedPacket[totalBytes] = eot;
			totalBytes ++;

			//slice the temp buffer into the final packet to send
			var packet = stuffedPacket.slice(0, totalBytes);

			//send
			port1.write(packet);
			console.log('Sending port1 ASCP Rate: ' + newRate.toString(10));
			
			//console.log('total bytes: ' + totalBytes);
			//console.log('packet length: ' + packet.length);
			//console.log('packet bytes: ' + packet.toString('hex'));
		}
		else if (port1Protocol == 'VDSTAD'){
			//first convert the rate value to a string
			var rateString;
			
			//vaderstad can only do up to 999.0, so cap it to this
			if (newRate > 999)
			{
				rateString = '999.0';
			}
			//else force to one decimal place and convert to string
			else
			{
				newRate = newRate.toFixed(1);
				rateString = newRate.toString(10);
			}
			
			//console.log('rate string: ' + rateString);
			
			//15 is max possible chars required
			var packet = new Buffer(15, 0);
			
			//vaderstad controller reads ascii characters
			//start of frame is ascii $
			packet[0] = 0x24;
			//set rate command is ascii SET followed by comma
			packet[1] = 0x53;
			packet[2] = 0x45;
			packet[3] = 0x54;
			packet[4] = 0x2C;
			
			//the rate value will have a min of 3 bytes, example 1.0
			packet[5] = Number(rateString.charCodeAt(0)).toString(10) & 0xFF;
			packet[6] = Number(rateString.charCodeAt(1)).toString(10) & 0xFF;
			packet[7] = Number(rateString.charCodeAt(2)).toString(10) & 0xFF;
			
			//but possibly 5 bytes long, example 100.0
			var pointer = 7;
			
			if (rateString.length > 3){
				pointer ++;
				packet[pointer] = Number(rateString.charCodeAt(3)).toString(10) & 0xFF;
			}
			
			if (rateString.length > 4){
				pointer ++;
				packet[pointer] = Number(rateString.charCodeAt(4)).toString(10) & 0xFF;
			}
			
			//end of frame byte is ascii *
			pointer ++;
			packet[pointer] = 0x2A;
			
			//checksum calc, xor of all bytes
			//exclude start of frame byte, exclude end of frame byte
			var chkSumValue = 0;
			
			chkSumValue = packet[1] ^ packet[2];
			chkSumValue = chkSumValue ^ packet[3];
			chkSumValue = chkSumValue ^ packet[4];
			chkSumValue = chkSumValue ^ packet[5];
			chkSumValue = chkSumValue ^ packet[6];
			chkSumValue = chkSumValue ^ packet[7];
			
			if (rateString.length > 3){
				chkSumValue = chkSumValue ^ packet[8];
			}
			if (rateString.length > 4){
				chkSumValue = chkSumValue ^ packet[9];
			}
			
			//mask to one byte
			chkSumValue = chkSumValue & 0xFF;
			
			//now convert the hex byte to a string, and set letters to upper case
			chkSumValue = chkSumValue.toString(16);
			chkSumValue = chkSumValue.toUpperCase();
			//console.log('checksum hex string: ' + chkSumValue);
			
			//then send the hex character code for the first character as one byte
			pointer ++;
			packet[pointer] = Number(chkSumValue.charCodeAt(0));
			//console.log('ascii code for first checksum character: ' + packet[pointer].toString(16));
			
			//and the same for the second character
			pointer ++;
			packet[pointer] = Number(chkSumValue.charCodeAt(1));
			//console.log('ascii code for second checksum character: ' + packet[pointer].toString(16));
			
			
			//end of frame is carriage return then line feed
			pointer ++;
			packet[pointer] = 0x0D;
			pointer ++;
			packet[pointer] = 0x0A;
			
			//slice the packet, up to and including the pointer
			//this drops any unused bytes at the end of the buffer
			pointer ++;
			packet = packet.slice(0, pointer);
			
			//console.log(packet);
			//console.log('packet bytes: ' + packet.toString('hex'));
			
			//send
			port1.write(packet);
			console.log('Sending port1 VDSTAD Rate: ' + rateString);
		}
		else if (port1Protocol == 'TEEJET844'){
			// Only difference between Teejet protocols is that the carrier application rate is limited to 999 for the 844 controller and to 9999 for the 855 controller.
			var rawRate;
                
			// Teejet 844 can only do up to 999, so cap it to this
			if (newRate > 999)
			{
				rawRate = 999;
			}
			// Only take integers, so remove any decimal places
			else
			{
				rawRate = newRate.toFixed(0);
			}
			var logRate = rawRate;
		
			// Packet is 15 bytes long
			// Bytes from position 0 to 11 are unused. Bytes in position 12 and 13 are rate. Byte in position 14 is checksum.
			// All are in Hexadecimal format.
			// When 2 bytes are used, the first received is the most significant byte.
		
			var packet = Buffer.alloc(15);
		
			packet[13] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[12] = rawRate & 0xFF;
		
		
			// Checksum is calculated by adding all bytes (without carry) and then substracting the result from 55 hexadecimal.
			var chkSumPointer = 0;
			var chkSumValue = 0;
		
			while (chkSumPointer < 14)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			chkSumValue = 0x55 - chkSumValue;
			packet[14] = chkSumValue & 0xFF;
		
			
			//send
			port1.write(packet);
			console.log('Sending port1 TEEJET844 Rate: ' + logRate);
		}
		else if (port1Protocol == 'TEEJET855'){
			// Only difference between Teejet protocols is that the carrier application rate is limited to 999 for the 844 controller and to 9999 for the 855 controller.
			var rawRate;
                
			// Teejet 855 can only do up to 9999, so cap it to this
			if (newRate > 9999)
			{
				rawRate = 9999;
			}
			// Only take integers, so remove any decimal places
			else
			{
				rawRate = newRate.toFixed(0);
			}
			var logRate = rawRate;
		
			// Packet is 15 bytes long
			// Bytes from position 0 to 11 are unused. Bytes in position 12 and 13 are rate. Byte in position 14 is checksum.
			// All are in Hexadecimal format.
			// When 2 bytes are used, the first received is the most significant byte.
		
			var packet = Buffer.alloc(15);
		
			packet[13] = rawRate & 0xFF;
			rawRate = rawRate >> 8;
			packet[12] = rawRate & 0xFF;
		
		
			// Checksum is calculated by adding all bytes (without carry) and then substracting the result from 55 hexadecimal.
			var chkSumPointer = 0;
			var chkSumValue = 0;
		
			while (chkSumPointer < 14)
			{
				chkSumValue += packet[chkSumPointer];
				chkSumPointer ++;
			}
			chkSumValue = 0x55 - chkSumValue;
			packet[14] = chkSumValue & 0xFF;
		
			
			//send
			port1.write(packet);
			console.log('Sending port1 TEEJET855 Rate: ' + logRate);
		}
	}
	catch (e){
		console.log('Catch error port1SendRate');
	}
}





//__________________________________________________________________________________________
//server for webpages, sockets, and http requests

var express = require('express');  
var app = express();
var server = require('http').createServer(app);
var io = require('socket.io')(server);

//when express receives a get request, send the index file
app.use(express.static(__dirname)); 
app.get('/', function(request, response, next) 
{	
	console.log('Received http get request');
    response.sendFile(__dirname + '/index.html');
});


//when express receives a post request, extract the data and respond
app.use(express.json());
app.use(express.urlencoded({ extended: true })); 
app.post('/', function(request, response, next) 
{	
	var bodyData = request.body;
	
	if (bodyData.command == 'refreshData'){
		
		ledGreen.digitalWrite(1);
		setTimeout(function(){ ledGreen.digitalWrite(0); }, 100);
		
		
		//console.log('Http post request: ' + JSON.stringify(bodyData));
		
		//parse values from the body data
		var ipadTargetRate = parseFloat(bodyData.targetRate);
		
		if (port0Mode == 'Rate Out'){
			port0SendRate(ipadTargetRate);
		}
		if (port1Mode == 'Rate Out'){
			port1SendRate(ipadTargetRate);
		}
	}
	
	//send new data to to ipad
	response.send({
		
	});
});


//__________________________________________________________________________________________
//webpage sockets

io.on('connection', function(socket) 
{  
	console.log('Socket is connected');
	
	socket.on('join', function(data) 
	{
		console.log(data);	
	});
	
	//__________________________________________________________________________________________
	//index webpage
	
	//webpage requesting data
	socket.on('indexPage', function(){
		socket.emit('newData',{
			port0Mode: port0Mode,
			port0Protocol: port0Protocol,
			port0Baud: port0Baud,
			port0DispConfig: port0DispConfig,
			port0FlowCtrl: port0FlowCtrl,
			port1Mode: port1Mode,
			port1Protocol: port1Protocol,
			port1Baud: port1Baud,
			port1DispConfig: port1DispConfig,
			port1FlowCtrl: port1FlowCtrl
		});
	});
	
	//cycle the mode
	socket.on('setPort0Mode', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		if (port0Mode == 'Off'){
			port0Mode = 'Rate Out';
		}
		else if (port0Mode == 'Rate Out'){
			port0Mode = 'Off';
		}
	});
	
	//cycle the protocol
	socket.on('setPort0Protocol', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		if (port0Protocol == 'LH5000'){
			port0Protocol = 'LH5000x10';
		}
		else if (port0Protocol == 'LH5000x10'){
			port0Protocol = 'LH5000/10';
		}
		else if (port0Protocol == 'LH5000/10'){
			port0Protocol = 'ZURF';
		}
		else if (port0Protocol == 'ZURF'){
			port0Protocol = 'ASCP';
		}
		else if (port0Protocol == 'ASCP'){
			port0Protocol = 'VDSTAD';
		}
		else if (port0Protocol == 'VDSTAD'){
			port0Protocol = 'TEEJET844';
		}
		else if (port0Protocol == 'TEEJET844'){
			port0Protocol = 'TEEJET855';
		}
		else if (port0Protocol == 'TEEJET855'){
			port0Protocol = 'LH5000';
		}

	});
	
	//cycle baud
	socket.on('setPort0Baud', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		switch (port0Baud){
			
			case 4800:
				port0Baud = 9600;
			break;
			
			case 9600:
				port0Baud = 19200;
			break;
			
			case 19200:
				port0Baud = 38400;
			break;
			
			case 38400:
				port0Baud = 57600;
			break;
			
			case 57600:
				port0Baud = 115200;
			break;
			
			case 115200:
				port0Baud = 4800;
			break;
			
			default:
		}
		
		//update the baud, see node serialport docs
		port0.update({
			baudRate: port0Baud
		});
	});
	
	//cycle the config
	socket.on('setPort0Config', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		//implement this if required
	});
	
	
	//cycle flow ctrl
	socket.on('setPort0FlowCtrl', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		if (port0FlowCtrl){
			port0FlowCtrl = false;
		}
		else {
			port0FlowCtrl = true;
		}
	});
	
	
	//cycle the mode
	socket.on('setPort1Mode', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		if (port1Mode == 'Off'){
			port1Mode = 'Rate Out';
		}
		else if (port1Mode == 'Rate Out'){
			port1Mode = 'Off';
		}
	});
	
	//cycle the protocol
	socket.on('setPort1Protocol', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		if (port1Protocol == 'LH5000'){
			port1Protocol = 'LH5000x10';
		}
		else if (port1Protocol == 'LH5000x10'){
			port1Protocol = 'LH5000/10';
		}
		else if (port1Protocol == 'LH5000/10'){
			port1Protocol = 'ZURF';
		}
		else if (port1Protocol == 'ZURF'){
			port1Protocol = 'ASCP';
		}
		else if (port1Protocol == 'ASCP'){
			port1Protocol = 'VDSTAD';
		}
		else if (port1Protocol == 'VDSTAD'){
			port1Protocol = 'TEEJET844';
		}
		else if (port1Protocol == 'TEEJET844'){
			port1Protocol = 'TEEJET855';
		}
		else if (port1Protocol == 'TEEJET855'){
			port1Protocol = 'LH5000';
		}

	});
	
	//cycle baud
	socket.on('setPort1Baud', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		switch (port1Baud){
			
			case 4800:
				port1Baud = 9600;
			break;
			
			case 9600:
				port1Baud = 19200;
			break;
			
			case 19200:
				port1Baud = 38400;
			break;
			
			case 38400:
				port1Baud = 57600;
			break;
			
			case 57600:
				port1Baud = 115200;
			break;
			
			case 115200:
				port1Baud = 4800;
			break;
			
			default:
		}
		
		//update the baud, see node serialport docs
		port1.update({
			baudRate: port1Baud
		});
	});
	
	//cycle the config
	socket.on('setPort1Config', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		//implement this if required
	});
	
	
	//cycle flow ctrl
	socket.on('setPort1FlowCtrl', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		if (port1FlowCtrl){
			port1FlowCtrl = false;
		}
		else {
			port1FlowCtrl = true;
		}
	});
	
	//__________________________________________________________________________________________
	//checks webpage
	
	//webpage requesting data
	socket.on('checksPage', function(){
		socket.emit('newData',{
			appRateCommand: appRateCommand,
			port0TxCounts: port0TxCounts,
			port0RxCounts: port0RxCounts,
			port1TxCounts: port1TxCounts,
			port1RxCounts: port1RxCounts
		});
	});
	
	socket.on('port0Rate', function (data){
		//set the new rate
		if (port0Mode == 'Rate Out'){
			port0SendRate(data);
		}
	});
	
	socket.on('port1Rate', function (data){
		//set the new rate
		if (port1Mode == 'Rate Out'){
			port1SendRate(data);
		}
	});
	
	//__________________________________________________________________________________________
	//advanced webpage
	
	//reset to factory settings
	socket.on('factoryReset', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveConfigFile, 4000);
		
		loadFactoryConfig();
	});
	
	//__________________________________________________________________________________________
	//info webpage
	
	//webpage requesting data
	socket.on('infoPage', function(){
		socket.emit('newData',{
			hardwareVer: hardwareVer,
			softwareVer: softwareVer
		});
	});
	
	//__________________________________________________________________________________________
	
});

//server port
server.listen(80);
